/**
 * Copyright (C) 2008 Gauvain Pocentek <gpocentek@linutop.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <unistd.h> /* R_OK */

#include <glib/gstdio.h>

#include "ltop-image.h"
#include "debug.h"

#define LTOP_IMAGE_DEFAULT_TIMEOUT 10

struct _LtopImagePrivate {
    GdkPixbuf *pixbuf;

    gfloat timeout;
    guint  timeout_source;
};

static void ltop_image_media_init (LtopMediaInterface *iface);
static void ltop_image_class_init (LtopImageClass *class);
static void ltop_image_init       (LtopImage *image);
static void ltop_image_finalize   (GObject *object);

static void ltop_image_run              (LtopMedia *media);
static void ltop_image_attach_to_window (LtopMedia *media,
                                         GdkWindow *window);

G_DEFINE_TYPE_EXTENDED (LtopImage,
                        ltop_image,
                        G_TYPE_OBJECT,
                        0,
                        G_IMPLEMENT_INTERFACE (LTOP_TYPE_MEDIA,
                                               ltop_image_media_init));

/* interface init */
static void
ltop_image_media_init (LtopMediaInterface *iface)
{
    iface->run = ltop_image_run;
    iface->attach_to_window = ltop_image_attach_to_window;
}

static void
ltop_image_class_init (LtopImageClass *class)
{
    g_type_class_add_private (class, sizeof (LtopImagePrivate));

    GObjectClass *object_class = (GObjectClass *) class;
    object_class->finalize = ltop_image_finalize;
}

static void
ltop_image_init (LtopImage *image)
{
    image->priv = G_TYPE_INSTANCE_GET_PRIVATE (image,
                                               LTOP_TYPE_IMAGE,
                                               LtopImagePrivate);

    image->priv->timeout = LTOP_IMAGE_DEFAULT_TIMEOUT;
    image->priv->timeout_source = 0;
    image->priv->pixbuf = NULL;
}

static void
ltop_image_finalize (GObject *object)
{
    DBG("entering");
    LtopImage *image = LTOP_IMAGE (object);

    if (image->priv->timeout_source)
    {
        g_source_remove (image->priv->timeout_source);
        image->priv->timeout_source = 0;
    }
    if (image->priv->pixbuf)
        g_object_unref (G_OBJECT (image->priv->pixbuf));

    (*G_OBJECT_CLASS (ltop_image_parent_class)->finalize) (object);
    DBG("LtopImage finalized");
    DBG("leaving");
}

static gboolean
raise_timed_out (gpointer data)
{
    DBG("timeout");
    ltop_media_job_done (LTOP_MEDIA (data));

    return FALSE;
}

GdkPixbuf *
get_pixbuf (gchar *filename, gint style)
{
    GdkScreen *screen;
    GdkPixbuf *pxb, *final = NULL, *tmp = NULL;
    gint w, h, iw = 0, ih = 0;
    gint dx, dy, xo, yo;
    gint i, j;
    gdouble xscale, yscale;
    GdkInterpType interp;

    g_return_val_if_fail (filename != NULL, NULL);

    if (style == LTOP_IMAGE_STYLE_TILED
        || style == LTOP_IMAGE_STYLE_CENTERED)
    {
        interp = GDK_INTERP_NEAREST;
    } else {
        interp = GDK_INTERP_BILINEAR;
    }

    screen = gdk_screen_get_default ();
    w = gdk_screen_get_width (screen);
    h = gdk_screen_get_height (screen);

    pxb = gdk_pixbuf_new_from_file (filename, NULL);
    if (!pxb)
        return NULL;

    iw = gdk_pixbuf_get_width (pxb);
    ih = gdk_pixbuf_get_height (pxb);

    /* set a black background */
    /* TODO: make the background color optional */
    final = gdk_pixbuf_new (GDK_COLORSPACE_RGB, FALSE, 8, w, h);
    gdk_pixbuf_fill (final, 0);

    /* Now the image itself */
    switch (style) {
    case LTOP_IMAGE_STYLE_CENTERED:
        dx = MAX((w - iw) / 2, 0);
        dy = MAX((h - ih) / 2, 0);
        xo = MIN((w - iw) / 2, dx);
        yo = MIN((h - ih) / 2, dy);

        gdk_pixbuf_composite (pxb, final, dx, dy,
            MIN (w, iw), MIN (h, ih), xo, yo, 1.0, 1.0,
            interp, 255);
        break;

    case LTOP_IMAGE_STYLE_TILED:
        tmp = gdk_pixbuf_new (GDK_COLORSPACE_RGB, TRUE, 8, w, h);
        for(i = 0; (i * iw) < w; i++) {
            for(j = 0; (j * ih) < h; j++) {
                gint newx = iw * i, newy = ih * j;
                gint neww = iw, newh = ih;

                if ((newx + neww) > w)
                    neww = w - newx;
                if ((newy + newh) > h)
                    newh = h - newy;

                gdk_pixbuf_copy_area (pxb, 0, 0,
                    neww, newh, tmp, newx, newy);
            }
        }

        gdk_pixbuf_composite (tmp, final, 0, 0, w, h,
            0, 0, 1.0, 1.0, interp, 255);
        g_object_unref (G_OBJECT (tmp));
        break;

    case LTOP_IMAGE_STYLE_STRETCHED:
        xscale = (gdouble) w / iw;
        yscale = (gdouble) h / ih;
        gdk_pixbuf_composite (pxb, final, 0, 0, w, h,
            0, 0, xscale, yscale, interp, 255);
        break;

    case LTOP_IMAGE_STYLE_SCALED:
        xscale = (gdouble) w / iw;
        yscale = (gdouble) h / ih;
        if (xscale < yscale) {
            yscale = xscale;
            xo = 0;
            yo = (h - (ih * yscale)) / 2;
        } else {
            xscale = yscale;
            xo = (w - (iw * xscale)) / 2;
            yo = 0;
        }
        dx = xo;
        dy = yo;

        gdk_pixbuf_composite (pxb, final, dx, dy,
            iw * xscale, ih * yscale, xo, yo, xscale, yscale,
            interp, 255);
        break;

    default:
        g_critical("Invalid image style: %d\n", (gint) style);
    }

    g_object_unref (pxb);

    return final;
}

GObject *
ltop_image_new (gchar *filename, gfloat timeout, gint style)
{
    g_return_val_if_fail (filename != NULL, NULL);
    g_return_val_if_fail (!g_access (filename, R_OK), NULL);

    LtopImage *image = g_object_new (LTOP_TYPE_IMAGE, NULL);
    image->priv->pixbuf = get_pixbuf (filename, style);
    if (timeout)
        image->priv->timeout = timeout;

    DBG("creating image: uri=%s timeout=%f style=%d",
        filename, image->priv->timeout, style);

    return G_OBJECT (image);
}

GObject *
ltop_image_new_black (gfloat timeout)
{
    GdkScreen *screen;
    int w, h;
    GdkPixbuf *pxb = NULL;

    LtopImage *image = g_object_new (LTOP_TYPE_IMAGE, NULL);

    screen = gdk_screen_get_default ();
    w = gdk_screen_get_width (screen);
    h = gdk_screen_get_height (screen);

    pxb = gdk_pixbuf_new (GDK_COLORSPACE_RGB, FALSE, 8, w, h);
    gdk_pixbuf_fill (pxb, 0);

    image->priv->pixbuf = pxb;

    if (timeout)
        image->priv->timeout = timeout;

    return G_OBJECT (image);
}

static void
ltop_image_run (LtopMedia *media)
{
    g_return_if_fail (LTOP_IS_MEDIA (media));

    DBG("running image");

    LtopImage *image = LTOP_IMAGE (media);
    if (!image->priv->timeout_source)
    {
        DBG("setting timeout");
        image->priv->timeout_source = g_timeout_add_seconds (
            image->priv->timeout, raise_timed_out, (gpointer) image);
    }
}

static void
ltop_image_attach_to_window (LtopMedia *media,
                             GdkWindow *window)
{
    DBG("Entering");
    g_return_if_fail (LTOP_IS_MEDIA (media));
    g_return_if_fail (GDK_IS_WINDOW (window));
    LtopImage *image = LTOP_IMAGE (media);

    /* TODO (Jannis:) You'd probably want to use a fallback pixbuf or skip 
     * all LtopImages for which the GdkPixbuf* could not be loaded. The
     * following two lines avoid segfaults when the pixbuf is NULL */
    if (image->priv->pixbuf == NULL)
        return;

    int h, w, x, y;
    w = gdk_pixbuf_get_width (image->priv->pixbuf);
    h = gdk_pixbuf_get_height (image->priv->pixbuf);

    gdk_error_trap_push ();

    gdk_window_get_origin (window, &x, &y);

    GdkRectangle rectangle = { x, y, w, h};

    gdk_window_invalidate_rect (window, &rectangle, FALSE);
    gdk_window_begin_paint_rect (window, &rectangle);
    gdk_draw_pixbuf (GDK_DRAWABLE (window),
                     gdk_gc_new (GDK_DRAWABLE (window)),
                     image->priv->pixbuf,
                     0, 0,
                     0, 0,
                     -1, -1,
                     GDK_RGB_DITHER_NORMAL,
                     0, 0);
    DBG("image attached to gdkwindow");
    gdk_window_end_paint (window);

    DBG("gdk_window_process_all_updates");
    gdk_window_process_all_updates ();

    gdk_flush ();
    gdk_error_trap_pop ();
}

